/*
 * kgraft_patch_bsc953052
 *
 * Fix for CVE-2015-7990, bsc#953052
 *
 *  SLE12 commit:
 *  e6b2c5897875561757c96b5c660c881d518b18f9
 *
 *  Copyright (c) 2015 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/socket.h>
#include <linux/net.h>
#include <linux/rds.h>
#include <net/sock.h>
#include <rdma/rdma_cm.h>

#include "local.h"

static struct rds_message *(*kgr_rds_message_alloc)(unsigned int extra_len, gfp_t gfp);
static int (*kgr_rds_rdma_extra_size)(struct rds_rdma_args *args);
static struct scatterlist *(*kgr_rds_message_alloc_sgs)(struct rds_message *rm, int nents);
static int (*kgr_rds_message_copy_from_user)(struct rds_message *rm, struct iovec *first_iov, size_t total_len);
static struct rds_connection *(*kgr_rds_conn_create_outgoing)(__be32 laddr, __be32 faddr, struct rds_transport *trans, gfp_t gfp);
static int (*kgr_rds_cmsg_rdma_args)(struct rds_sock *rs, struct rds_message *rm, struct cmsghdr *cmsg);
static int (*kgr_rds_cmsg_rdma_dest)(struct rds_sock *rs, struct rds_message *rm, struct cmsghdr *cmsg);
static int (*kgr_rds_cmsg_rdma_map)(struct rds_sock *rs, struct rds_message *rm, struct cmsghdr *cmsg);
static int (*kgr_rds_cmsg_atomic)(struct rds_sock *rs, struct rds_message *rm, struct cmsghdr *cmsg);
static void (*kgr_rds_conn_connect_if_down)(struct rds_connection *conn);
static int (*kgr_rds_cong_wait)(struct rds_cong_map *map, __be16 port, int nonblock, struct rds_sock *rs);
static int (*kgr_rds_send_queue_rm)(struct rds_sock *rs, struct rds_connection *conn, struct rds_message *rm, __be16 sport, __be16 dport, int *queued);
static int (*kgr_rds_send_xmit)(struct rds_connection *conn);
static void (*kgr_rds_message_put)(struct rds_message *rm);
static void (*kgr_rds_rdma_unuse)(struct rds_sock *rs, u32 r_key, int force);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "rds_message_alloc", (void *)&kgr_rds_message_alloc },
	{ "rds_rdma_extra_size", (void *)&kgr_rds_rdma_extra_size },
	{ "rds_message_alloc_sgs", (void *)&kgr_rds_message_alloc_sgs },
	{ "rds_message_copy_from_user", (void *)&kgr_rds_message_copy_from_user },
	{ "rds_conn_create_outgoing", (void *)&kgr_rds_conn_create_outgoing },
	{ "rds_cmsg_rdma_args", (void *)&kgr_rds_cmsg_rdma_args },
	{ "rds_cmsg_rdma_dest", (void *)&kgr_rds_cmsg_rdma_dest },
	{ "rds_cmsg_rdma_map", (void *)&kgr_rds_cmsg_rdma_map },
	{ "rds_cmsg_atomic", (void *)&kgr_rds_cmsg_atomic },
	{ "rds_conn_connect_if_down", (void *)&kgr_rds_conn_connect_if_down },
	{ "rds_cong_wait", (void *)&kgr_rds_cong_wait },
	{ "rds_send_queue_rm", (void *)&kgr_rds_send_queue_rm },
	{ "rds_send_xmit", (void *)&kgr_rds_send_xmit },
	{ "rds_message_put", (void *)&kgr_rds_message_put },
	{ "rds_rdma_unuse", (void *)&kgr_rds_rdma_unuse },
};

static int kgr_rds_rm_size(struct msghdr *msg, int data_len)
{
	struct cmsghdr *cmsg;
	int size = 0;
	int cmsg_groups = 0;
	int retval;

	for (cmsg = CMSG_FIRSTHDR(msg); cmsg; cmsg = CMSG_NXTHDR(msg, cmsg)) {
		if (!CMSG_OK(msg, cmsg))
			return -EINVAL;

		if (cmsg->cmsg_level != SOL_RDS)
			continue;

		switch (cmsg->cmsg_type) {
		case RDS_CMSG_RDMA_ARGS:
			cmsg_groups |= 1;
			retval = kgr_rds_rdma_extra_size(CMSG_DATA(cmsg));
			if (retval < 0)
				return retval;
			size += retval;

			break;

		case RDS_CMSG_RDMA_DEST:
		case RDS_CMSG_RDMA_MAP:
			cmsg_groups |= 2;
			/* these are valid but do no add any size */
			break;

		case RDS_CMSG_ATOMIC_CSWP:
		case RDS_CMSG_ATOMIC_FADD:
		case RDS_CMSG_MASKED_ATOMIC_CSWP:
		case RDS_CMSG_MASKED_ATOMIC_FADD:
			cmsg_groups |= 1;
			size += sizeof(struct scatterlist);
			break;

		default:
			return -EINVAL;
		}

	}

	size += ceil(data_len, PAGE_SIZE) * sizeof(struct scatterlist);

	/* Ensure (DEST, MAP) are never used with (ARGS, ATOMIC) */
	if (cmsg_groups == 3)
		return -EINVAL;

	return size;
}

static int kgr_rds_cmsg_send(struct rds_sock *rs, struct rds_message *rm,
			 struct msghdr *msg, int *allocated_mr)
{
	struct cmsghdr *cmsg;
	int ret = 0;

	for (cmsg = CMSG_FIRSTHDR(msg); cmsg; cmsg = CMSG_NXTHDR(msg, cmsg)) {
		if (!CMSG_OK(msg, cmsg))
			return -EINVAL;

		if (cmsg->cmsg_level != SOL_RDS)
			continue;

		/* As a side effect, RDMA_DEST and RDMA_MAP will set
		 * rm->rdma.m_rdma_cookie and rm->rdma.m_rdma_mr.
		 */
		switch (cmsg->cmsg_type) {
		case RDS_CMSG_RDMA_ARGS:
			ret = kgr_rds_cmsg_rdma_args(rs, rm, cmsg);
			break;

		case RDS_CMSG_RDMA_DEST:
			ret = kgr_rds_cmsg_rdma_dest(rs, rm, cmsg);
			break;

		case RDS_CMSG_RDMA_MAP:
			ret = kgr_rds_cmsg_rdma_map(rs, rm, cmsg);
			if (!ret)
				*allocated_mr = 1;
			break;
		case RDS_CMSG_ATOMIC_CSWP:
		case RDS_CMSG_ATOMIC_FADD:
		case RDS_CMSG_MASKED_ATOMIC_CSWP:
		case RDS_CMSG_MASKED_ATOMIC_FADD:
			ret = kgr_rds_cmsg_atomic(rs, rm, cmsg);
			break;

		default:
			return -EINVAL;
		}

		if (ret)
			break;
	}

	return ret;
}

/* Patched function */
int kgr_rds_sendmsg(struct kiocb *iocb, struct socket *sock, struct msghdr *msg,
		size_t payload_len)
{
	struct sock *sk = sock->sk;
	struct rds_sock *rs = rds_sk_to_rs(sk);
	struct sockaddr_in *usin = (struct sockaddr_in *)msg->msg_name;
	__be32 daddr;
	__be16 dport;
	struct rds_message *rm = NULL;
	struct rds_connection *conn;
	int ret = 0;
	int queued = 0, allocated_mr = 0;
	int nonblock = msg->msg_flags & MSG_DONTWAIT;
	long timeo = sock_sndtimeo(sk, nonblock);

	/* Mirror Linux UDP mirror of BSD error message compatibility */
	/* XXX: Perhaps MSG_MORE someday */
	if (msg->msg_flags & ~(MSG_DONTWAIT | MSG_CMSG_COMPAT)) {
		ret = -EOPNOTSUPP;
		goto out;
	}

	if (msg->msg_namelen) {
		/* XXX fail non-unicast destination IPs? */
		if (msg->msg_namelen < sizeof(*usin) || usin->sin_family != AF_INET) {
			ret = -EINVAL;
			goto out;
		}
		daddr = usin->sin_addr.s_addr;
		dport = usin->sin_port;
	} else {
		/* We only care about consistency with ->connect() */
		lock_sock(sk);
		daddr = rs->rs_conn_addr;
		dport = rs->rs_conn_port;
		release_sock(sk);
	}

	/*
	 * Fix CVE-2015-7990
	 */
	lock_sock(sk);
	if (daddr == 0 || rs->rs_bound_addr == 0) {
		release_sock(sk);
		ret = -ENOTCONN; /* XXX not a great errno */
		goto out;
	}
	release_sock(sk);

	/* size of rm including all sgs */
	ret = kgr_rds_rm_size(msg, payload_len);
	if (ret < 0)
		goto out;

	rm = kgr_rds_message_alloc(ret, GFP_KERNEL);
	if (!rm) {
		ret = -ENOMEM;
		goto out;
	}

	/* Attach data to the rm */
	if (payload_len) {
		rm->data.op_sg = kgr_rds_message_alloc_sgs(rm, ceil(payload_len, PAGE_SIZE));
		if (!rm->data.op_sg) {
			ret = -ENOMEM;
			goto out;
		}
		ret = kgr_rds_message_copy_from_user(rm, msg->msg_iov, payload_len);
		if (ret)
			goto out;
	}
	rm->data.op_active = 1;

	rm->m_daddr = daddr;

	/* rds_conn_create has a spinlock that runs with IRQ off.
	 * Caching the conn in the socket helps a lot. */
	if (rs->rs_conn && rs->rs_conn->c_faddr == daddr)
		conn = rs->rs_conn;
	else {
		conn = kgr_rds_conn_create_outgoing(rs->rs_bound_addr, daddr,
					rs->rs_transport,
					sock->sk->sk_allocation);
		if (IS_ERR(conn)) {
			ret = PTR_ERR(conn);
			goto out;
		}
		rs->rs_conn = conn;
	}

	/* Parse any control messages the user may have included. */
	ret = kgr_rds_cmsg_send(rs, rm, msg, &allocated_mr);
	if (ret)
		goto out;

	if (rm->rdma.op_active && !conn->c_trans->xmit_rdma) {
		printk_ratelimited(KERN_NOTICE "rdma_op %p conn xmit_rdma %p\n",
			       &rm->rdma, conn->c_trans->xmit_rdma);
		ret = -EOPNOTSUPP;
		goto out;
	}

	if (rm->atomic.op_active && !conn->c_trans->xmit_atomic) {
		printk_ratelimited(KERN_NOTICE "atomic_op %p conn xmit_atomic %p\n",
			       &rm->atomic, conn->c_trans->xmit_atomic);
		ret = -EOPNOTSUPP;
		goto out;
	}

	kgr_rds_conn_connect_if_down(conn);

	ret = kgr_rds_cong_wait(conn->c_fcong, dport, nonblock, rs);
	if (ret) {
		rs->rs_seen_congestion = 1;
		goto out;
	}

	while (!kgr_rds_send_queue_rm(rs, conn, rm, rs->rs_bound_port,
				  dport, &queued)) {
		/*
		 * KGR - no rds_stats due to missing percpu in kallsyms
		 * -rds_stats_inc(s_send_queue_full);
		 */
		/* XXX make sure this is reasonable */
		if (payload_len > rds_sk_sndbuf(rs)) {
			ret = -EMSGSIZE;
			goto out;
		}
		if (nonblock) {
			ret = -EAGAIN;
			goto out;
		}

		timeo = wait_event_interruptible_timeout(*sk_sleep(sk),
					kgr_rds_send_queue_rm(rs, conn, rm,
							  rs->rs_bound_port,
							  dport,
							  &queued),
					timeo);
		/*
		 * KGR - just debug call, we have no DEBUG
		 * -rdsdebug("sendmsg woke queued %d timeo %ld\n", queued, timeo);
		 */
		if (timeo > 0 || timeo == MAX_SCHEDULE_TIMEOUT)
			continue;

		ret = timeo;
		if (ret == 0)
			ret = -ETIMEDOUT;
		goto out;
	}

	/*
	 * By now we've committed to the send.  We reuse rds_send_worker()
	 * to retry sends in the rds thread if the transport asks us to.
	 */
	/*
	 * KGR - no rds_stats due to missing percpu in kallsyms
	 * -rds_stats_inc(s_send_queued);
	 */

	if (!test_bit(RDS_LL_SEND_FULL, &conn->c_flags))
		kgr_rds_send_xmit(conn);

	kgr_rds_message_put(rm);
	return payload_len;

out:
	/* If the user included a RDMA_MAP cmsg, we allocated a MR on the fly.
	 * If the sendmsg goes through, we keep the MR. If it fails with EAGAIN
	 * or in any other way, we need to destroy the MR again */
	if (allocated_mr)
		kgr_rds_rdma_unuse(rs, rds_rdma_cookie_key(rm->m_rdma_cookie), 1);

	if (rm)
		kgr_rds_message_put(rm);
	return ret;
}

static int kgr_patch_bsc953052_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	pr_warn("kgraft-patch: due to a patch implementation rds_stats can be distorted from now on\n");

	return 0;
}

static int kgr_patch_bsc953052_module_notify(struct notifier_block *nb,
                unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, "rds"))
		return 0;

	ret = kgr_patch_bsc953052_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc953052_module_nb = {
	.notifier_call = kgr_patch_bsc953052_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc953052_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module("rds")) {
		ret = kgr_patch_bsc953052_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc953052_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc953052_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc953052_module_nb);
}
